<!--
	Copyright 2009 The Go Authors. All rights reserved.
	Use of this source code is governed by a BSD-style
	license that can be found in the LICENSE file.
-->
{{with .Alert}}
	<p>
	<span class="alert" style="font-size:120%">{{html .}}</span>
	</p>
{{end}}
{{with .Alt}}
	<p>
	<span class="alert" style="font-size:120%">Did you mean: </span>
	{{range .Alts}}
		<a href="search?q={{urlquery .}}" style="font-size:120%">{{html .}}</a>
	{{end}}
	</p>
{{end}}
